#include <avr\io.h>	
#include "mcp.h"
#include "mcp2515_can.h"

#define	DDR_CS		DDRB
#define	PORT_CS		PORTB
#define	DDR_SPI		DDRB
#define	PORT_SPI	PORTB
#define	SPI_CS		4
#define	SPI_MISO	6
#define	SPI_MOSI	5
#define	SPI_SCK		7
#define F_CPU		4000000

//SPI Prototypen:
void spi_init(void);
uint8_t spi_putc(uint8_t data);
//MCP Prototypen:
void mcp2515_write_register( uint8_t adress, uint8_t data );
uint8_t mcp2515_read_register(uint8_t adress);
void mcp2515_bit_modify(uint8_t adress, uint8_t mask, uint8_t data);
void mcp2515_init(void);
void delay(uint16_t ms);
uint8_t can_send_message(CANMessage *p_message);


//____________________________________________________________________________________
//SPI-Funktionen:
void spi_init(void)
{
	// Aktivieren der Pins fr das SPI Interface
	DDR_SPI	 |= (1<<SPI_SCK)|(1<<SPI_MOSI);
	PORT_SPI &= ~((1<<SPI_SCK)|(1<<SPI_MOSI)|(1<<SPI_MISO));

	DDR_CS	 |= (1<<SPI_CS);
	PORT_CS  |= (1<<SPI_CS);

	// Aktivieren des SPI Master Interfaces, fosc = fclk / 16
	SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0);
}
//____________________________________________________________________________________

uint8_t spi_putc(uint8_t data)
{
	// Sendet ein uint8_t
	SPDR = data;

	// Wartet bis uint8_t gesendet wurde
	while( !( SPSR & (1<<SPIF) ) );

	return SPDR;
}
//____________________________________________________________________________________

void mcp2515_write_register( uint8_t adress, uint8_t data )
{
	// /CS des MCP2515 auf Low ziehen
	PORT_CS &= ~(1<<SPI_CS);

	spi_putc(SPI_WRITE);
	spi_putc(adress);
	spi_putc(data);

	// /CS Leitung wieder freigeben
	PORT_CS |= (1<<SPI_CS);
}
//____________________________________________________________________________________

uint8_t mcp2515_read_register(uint8_t adress)
{
	uint8_t data;

	// /CS des MCP2515 auf Low ziehen
	PORT_CS &= ~(1<<SPI_CS);

	spi_putc(SPI_READ);
	spi_putc(adress);

	data = spi_putc(0xff);

	// /CS Leitung wieder freigeben
	PORT_CS |= (1<<SPI_CS);

	return data;
}

//____________________________________________________________________________________

void mcp2515_bit_modify(uint8_t adress, uint8_t mask, uint8_t data)
{
	// /CS des MCP2515 auf Low ziehen
	PORT_CS &= ~(1<<SPI_CS);

	spi_putc(SPI_BIT_MODIFY);
	spi_putc(adress);
	spi_putc(mask);
	spi_putc(data);

	// /CS Leitung wieder freigeben
	PORT_CS |= (1<<SPI_CS);
}


//____________________________________________________________________________________

void mcp2515_init(void)
{
	// SPI Interface initialisieren
	spi_init();
	delay(50);
	// MCP2515 per Software Reset zurcksetzten,
	// danach ist der MCP2515 im Configuration Mode
	PORTB &= ~(1<<SPI_CS);
	spi_putc( SPI_RESET );
	PORTB |= (1<<SPI_CS);
	delay(50);

	// Bit Timing, 125 kHz @ 4MHz OSC

	// BRP = 3
	mcp2515_write_register( CNF1, 0 );		//3

	// Prop Seg und Phase Seg einstellen
	mcp2515_write_register( CNF2, (1<<BTLMODE)|(1<<PHSEG11));	// 0xB8

	// Wake-up Filter deaktivieren, Phase Seg einstelen
	mcp2515_write_register( CNF3, (1<<PHSEG21));		// 0x05

	// Aktivieren der Rx Buffer Interrupts (am INT wird dann Interupt ausgelst)
	mcp2515_write_register( CANINTE, (1<<RX1IE)|(1<<RX0IE) );

	/*
	 *	Einstellen der Filter
	 */

	// Buffer 0 : Empfangen aller Nachrichten
	mcp2515_write_register( RXB0CTRL, (1<<RXM1)|(1<<RXM0) );

	// Buffer 1 : Empfangen aller Nachrichten
	mcp2515_write_register( RXB1CTRL, (1<<RXM1)|(1<<RXM0) );

	// Alle Bits der Empfangsmaske lschen,
	// damit werden alle Nachrichten empfangen
	mcp2515_write_register( RXM0SIDH, 0 );
	mcp2515_write_register( RXM0SIDL, 0 );
	mcp2515_write_register( RXM0EID8, 0 );
	mcp2515_write_register( RXM0EID0, 0 );

	mcp2515_write_register( RXM1SIDH, 0 );
	mcp2515_write_register( RXM1SIDL, 0 );
	mcp2515_write_register( RXM1EID8, 0 );
	mcp2515_write_register( RXM1EID0, 0 );

	/*
	 *	Einstellen der Pin Funktionen
	 */

	// Deaktivieren der Pins RXnBF  ( auf High Impedance )
	mcp2515_write_register( BFPCTRL, 0 );

	// TXnRTS Bits als Inputs schalten
	mcp2515_write_register( TXRTSCTRL, 0 );

	// Device zurck in den normalen Modus versetzten
	mcp2515_bit_modify( CANCTRL, 0xE0, 0);
}

//____________________________________________________________________________________

void delay(uint16_t ms)
{
	uint16_t zaehler;
	
	while (ms) 
	{
		zaehler = F_CPU / 5000;
		while (zaehler) 
		{
			asm volatile ("nop");
			zaehler--;
		}
		ms--;
	}
}
//____________________________________________________________________________________

uint8_t can_send_message(CANMessage *p_message)
{
     uint8_t status, address;

     // Status des MCP2515 auslesen
     PORT_CS &= ~(1<<SPI_CS);
     spi_putc(SPI_READ_STATUS);
     status = spi_putc(0xff);
     spi_putc(0xff);
     PORT_CS |= (1<<SPI_CS);

     /* Statusbyte:
      *
      * Bit     Funktion
      *  2     TXB0CNTRL.TXREQ
      *  4     TXB1CNTRL.TXREQ
      *  6     TXB2CNTRL.TXREQ
      */

     if (bit_is_clear(status, 2)) {
          address = 0x00;
     }
     else if (bit_is_clear(status, 4)) {
          address = 0x02;
     }
     else if (bit_is_clear(status, 6)) {
          address = 0x04;
     }
     else {
          /* Alle Puffer sind belegt,
             Nachricht kann nicht verschickt werden */
          return 0;
     }

     PORT_CS &= ~(1<<SPI_CS);          // CS Low
     spi_putc(SPI_WRITE_TX | address);

     // Standard ID einstellen
     spi_putc((uint8_t) (p_message->id>>3));
     spi_putc((uint8_t) (p_message->id<<5));

     // Extended ID
     spi_putc(0x00);
     spi_putc(0x00);

     uint8_t length = p_message->length;

     if (length > 8) {
          length = 8;
     }

     // Ist die Nachricht ein "Remote Transmit Request" ?
     if (p_message->rtr)
     {
          /* Ein RTR hat zwar eine Lnge,
             aber enthlt keine Daten */

          // Nachrichten Lnge + RTR einstellen
          spi_putc((1<<RTR) | length);
     }
     else
     {
          // Nachrichten Lnge einstellen
          spi_putc(length);

          // Daten
		  uint8_t i;
          for (i=0;i<length;i++) {
               spi_putc(p_message->data[i]);
          }
     }
     PORT_CS |= (1<<SPI_CS);          // CS auf High

     asm volatile ("nop");

     /* CAN Nachricht verschicken
        die letzten drei Bit im RTS Kommando geben an welcher
        Puffer gesendet werden soll */
     PORT_CS &= ~(1<<SPI_CS);          // CS wieder Low
     if (address == 0x00) {
          spi_putc(SPI_RTS | 0x01);
     } else {
          spi_putc(SPI_RTS | address);
     }
     PORT_CS |= (1<<SPI_CS);          // CS auf High

     return 1;
}


